################################################################################
# UNIT TESTS
################################################################################

from pdstable import *

import unittest

class Test_Pds4Table(unittest.TestCase):

  def runTest(self):

    INDEX_PATH = 'test_files/uranus_occultations_index.xml'

    test_table_basic = PdsTable(INDEX_PATH)

    # Test strings
    test_file_names = test_table_basic.column_values['File Name']
    file_name_test_set = np.array([
                            'u0201_palomar_508cm_2200nm_radius_equator_egress_1000m.xml',
                            'u0201_palomar_508cm_2200nm_radius_equator_egress_100m.xml',
                            'u0201_palomar_508cm_2200nm_radius_equator_egress_500m.xml',
                            'u0201_palomar_508cm_2200nm_radius_equator_ingress_1000m.xml'
                         ])
    self.assertTrue(np.all(file_name_test_set == test_file_names[0:4]))

    # Test ints
    test_data_quality_idx = test_table_basic.column_values['Data Quality Index']
    data_quality_idx_test_set = np.array([6, 6, 6, 6])
    self.assertTrue(np.all(data_quality_idx_test_set == test_data_quality_idx[:4]))

    # Test floats
    test_proj_star_diameter = test_table_basic.column_values['Projected Star Diameter']
    proj_star_diameter_test_set = np.array([0.3, 0.3, 0.3, 0.3])
    self.assertTrue(np.all(proj_star_diameter_test_set == test_proj_star_diameter[:4]))

    # Test times as strings
    test_start_time_strs = test_table_basic.column_values['Start Time UTC']
    start_time_str_test_set = ['2002-07-29T10:03:17.0654Z', '2002-07-29T10:03:13.4075Z',
                               '2002-07-29T10:03:15.4524Z', '2002-07-29T09:48:38.6088Z']
    self.assertEqual(start_time_str_test_set[0], test_start_time_strs[0])
    self.assertEqual(start_time_str_test_set[1], test_start_time_strs[1])
    self.assertEqual(start_time_str_test_set[2], test_start_time_strs[2])
    self.assertEqual(start_time_str_test_set[3], test_start_time_strs[3])

    self.assertTrue(isinstance(test_start_time_strs, np.ndarray))
    self.assertTrue(isinstance(test_start_time_strs[0], np.str_))

    # Test dicts_by_row()
    rowdict = test_table_basic.dicts_by_row()
    for i in range(4):
        self.assertEqual(rowdict[i]['Start Time UTC'], test_start_time_strs[i])

    rowvals = test_table_basic.get_column('Start Time UTC')
    rowmasks = test_table_basic.get_column_mask('Start Time UTC')
    for i in range(10):
        self.assertEqual(rowdict[i]['Start Time UTC'], rowvals[i])
        self.assertFalse(rowmasks[i])

    ####################################
    # Test times as seconds (floats)
    ####################################

    test_table_secs = PdsTable(INDEX_PATH, times=['Start Time UTC'])

    test_start_times = test_table_secs.column_values['Start Time UTC']
    start_time_test_set = np.array([81209029.0654, 81209025.4075,
                                    81209027.4524, 81208150.6088])
    self.assertTrue(np.all(start_time_test_set == test_start_times[:4]))
    self.assertTrue(isinstance(start_time_test_set, np.ndarray))

    # Test dicts_by_row()
    rowdict = test_table_secs.dicts_by_row()
    for i in range(4):
        self.assertEqual(rowdict[i]['Start Time UTC'], start_time_test_set[i])

    rowvals = test_table_secs.get_column('Start Time UTC')
    rowmask = test_table_secs.get_column_mask('Start Time UTC')
    for i in range(10):
        self.assertEqual(rowdict[i]['Start Time UTC'], rowvals[i])
        self.assertFalse(rowmask[i])

    ####################################
    # Row lookups
    ####################################
    # File Specification
    self.assertEqual(test_table_basic.filespec_column_index(), 2)
    # Bundle Name
    self.assertEqual(test_table_basic.volume_column_index(), 4)
    # File Specification
    self.assertEqual(test_table_basic.find_row_index_by_volume_filespec(
            '',
            r'2021-04-uranus-redelivery\uranus_occ_u0201_palomar_508cm\data\global' +
            r'\u0201_palomar_508cm_2200nm_radius_equator_egress_500m.xml'), 2)
    self.assertEqual(test_table_basic.find_row_indices_by_volume_filespec(
            '',
            r'2021-04-uranus-redelivery\uranus_occ_u0201_palomar_508cm\data\global' +
            r'\u0201_palomar_508cm_2200nm_radius_equator_egress_500m.xml'), [2])
    self.assertEqual(test_table_basic.find_row_index_by_volume_filespec(
            '',
            r'2021-04-uranus-redelivery\uranus_occ_u0201_palomar_508cm\data\global' +
            r'\u0201_palomar_508cm_2200nm_radius_equator_ingress_1000m.xml'), 3)
    self.assertEqual(test_table_basic.find_row_indices_by_volume_filespec(
            '',
            r'2021-04-uranus-redelivery\uranus_occ_u0201_palomar_508cm\data\global' +
            r'\u0201_palomar_508cm_2200nm_radius_equator_ingress_1000m.xml'), [3])
    # Bundle Name & File Specification
    self.assertEqual(test_table_basic.find_row_index_by_volume_filespec(
            'uranus_occ_u0_kao_91cm',
            r'2021-04-uranus-redelivery\uranus_occ_u0_kao_91cm\data\global' +
            r'\u0_kao_91cm_734nm_radius_equator_egress_1000m.xml'), 21)
    self.assertEqual(test_table_basic.find_row_indices_by_volume_filespec(
            'uranus_occ_u0_kao_91cm',
            r'2021-04-uranus-redelivery\uranus_occ_u0_kao_91cm\data\global' +
            r'\u0_kao_91cm_734nm_radius_equator_egress_1000m.xml'), [21])

    ####################################
    # Row ranges
    ####################################

    partial_table = PdsTable(INDEX_PATH, row_range=(2,4))
    self.assertEqual(partial_table.rows, 2)

    self.assertEqual(partial_table.filespec_column_index(), 2)
    self.assertEqual(partial_table.volume_column_index(), 4)

    self.assertEqual(partial_table.find_row_index_by_volume_filespec(
            '',
            r'2021-04-uranus-redelivery\uranus_occ_u0201_palomar_508cm\data\global' +
            r'\u0201_palomar_508cm_2200nm_radius_equator_egress_500m.xml'), 0)
    self.assertEqual(partial_table.find_row_indices_by_volume_filespec(
            '',
            r'2021-04-uranus-redelivery\uranus_occ_u0201_palomar_508cm\data\global' +
            r'\u0201_palomar_508cm_2200nm_radius_equator_egress_500m.xml'), [0])

    self.assertEqual(partial_table.find_row_index_by_volume_filespec(
            '',
            r'2021-04-uranus-redelivery\uranus_occ_u0201_palomar_508cm\data\global' +
            r'\u0201_palomar_508cm_2200nm_radius_equator_ingress_1000m.xml'), 1)
    self.assertEqual(partial_table.find_row_indices_by_volume_filespec(
            '',
            r'2021-04-uranus-redelivery\uranus_occ_u0201_palomar_508cm\data\global' +
            r'\u0201_palomar_508cm_2200nm_radius_equator_ingress_1000m.xml'), [1])

    self.assertEqual(partial_table.find_row_index_by_volume_filespec(
            'uranus_occ_u0201_palomar_508cm',
            r'2021-04-uranus-redelivery\uranus_occ_u0201_palomar_508cm\data\global' +
            r'\u0201_palomar_508cm_2200nm_radius_equator_egress_500m.xml'), 0)
    self.assertEqual(partial_table.find_row_indices_by_volume_filespec(
            'uranus_occ_u0201_palomar_508cm',
            r'2021-04-uranus-redelivery\uranus_occ_u0201_palomar_508cm\data\global' +
            r'\u0201_palomar_508cm_2200nm_radius_equator_egress_500m.xml'), [0])

    ####################################
    # PdsLabel input option
    ####################################
    # For PDS4, we store the label dictionary in .lable instead of pdsparser.PdsLabel
    # instance, therefore we use "==" here instead of "is"
    test = PdsTable(INDEX_PATH, label_contents=partial_table.pdslabel)
    self.assertTrue(test.pdslabel == partial_table.pdslabel)

    # PDS4 TODO: Add tests for invalids & replacements



##########################################################################################
# pdstable/pds4table.py
# Store Pds4TableInfo and Pds4ColumnInfo
##########################################################################################
import julian
import numbers
import numpy as np
import os
import re

from collections import defaultdict
from pds4_tools.reader.label_objects import Label

PDS4_LBL_EXTENSIONS = {'.xml', '.lblx'}

def is_pds4_label(label_name):
    """Check if the given label is a PDS4 label."""

    for ext in PDS4_LBL_EXTENSIONS:
        if label_name.endswith(ext):
            return True

PDS4_BUNDLE_COLNAMES = (
    'Bundle Name',
)
# The mapping of a product tag to its corresponding file area tags
# The key is a product component tag, and the value is its corresponding file area tag,
# it could be just one (string) or multiple (tuple) file area tags
PDS4_PRODUCT_TO_FILE_AREA_TAGS_MAPPING = {
    'Product_Ancillary': 'File_Area_Ancillary',
    'Product_Browse': 'File_Area_Browse',
    'Product_Metadata_Supplemental': 'File_Area_Metadata',
    'Product_Observational': ('File_Area_Observational',
                              'File_Area_Observational_Supplemental'),
}

# The mapping of a table tag to its corresponding record and field tags
# The key is a table tag, and the value is a tuple of the record and field tags
PDS4_TABLE_TO_RECORD_FIELD_TAGS_MAPPING = {
    'Table_Binary': ('Record_Binary', 'Field_Binary'),
    'Table_Character': ('Record_Character', 'Field_Character'),
    'Table_Delimited': ('Record_Delimited', 'Field_Delimited'),
    'Table_Delimited_Source_Product_External': ('Record_Delimited', 'Field_Delimited'),
    'Table_Delimited_Source_Product_Internal': ('Record_Delimited', 'Field_Delimited'),
}

# PDS4 label tags under Special_Constants
PDS4_SPECIAL_CONSTANTS_TAGS = {
    'error_constant',
 	'high_instrument_saturation',
 	'high_representation_saturation',
 	'invalid_constant',
 	'low_instrument_saturation',
 	'low_representation_saturation',
 	'missing_constant',
 	'not_applicable_constant',
 	'saturated_constant',
 	'unknown_constant',
 	# 'valid_maximum',
 	# 'valid_minimum'
}

# This is an exhaustive tuple of string-like types
STRING_TYPES = (str, bytes, bytearray, np.str_, np.bytes_)

# Needed because the default value of strip is False
def tai_from_iso(string):
    return julian.tai_from_iso(string, strip=True)

def int_from_base2(string):
    return int(string, 2)

def int_from_base8(string):
    return int(string, 8)

def int_from_base16(string):
    return int(string, 16)

# Delimiter used to separate column values in the same row
# It's encoded by 'UTF-8'
PDS4_FIELD_DELIMITER = {
    'Carriage-Return Line-Feed': b'\r\n',
    'Comma': b',',
    'Horizontal Tab': b'\t',
    'Semicolon': b';',
    'Vertical Bar': b'|'
}

# key: PDS4 data type
# value: a tuple of (self.data_type, self.dtype2, self.scalar_func)
PDS4_CHR_DATA_TYPE_MAPPING = {
    'ASCII_Date_DOY': ('time', 'S', tai_from_iso),
    'ASCII_Date_Time_DOY': ('time', 'S', tai_from_iso),
    'ASCII_Date_Time_DOY_UTC': ('time', 'S', tai_from_iso),
    'ASCII_Date_Time_YMD': ('time', 'S', tai_from_iso),
    'ASCII_Date_Time_YMD_UTC': ('time', 'S', tai_from_iso),
    'ASCII_Date_YMD': ('time', 'S', tai_from_iso),
    'ASCII_Time': ('time', 'S', tai_from_iso),
    'ASCII_Integer': ('int', 'int', int),
    'ASCII_NonNegative_Integer': ('int', 'int', int),
    'ASCII_Real': ('float', 'float', float),
    'ASCII_AnyURI': ('string', 'U', None),
    'ASCII_Directory_Path_Name': ('string', 'U', None),
    'ASCII_DOI': ('string', 'U', None),
    'ASCII_File_Name': ('string', 'U', None),
    'ASCII_File_Specification_Name': ('string', 'U', None),
    'ASCII_LID': ('string', 'U', None),
    'ASCII_LIDVID': ('string', 'U', None),
    'ASCII_LIDVID_LID': ('string', 'U', None),
    'ASCII_MD5_Checksum': ('string', 'U', None),
    'ASCII_String': ('string', 'U', None),
    'ASCII_VID': ('string', 'U', None),
    'UTF8_String': ('string', 'U', None),
    'ASCII_Boolean': ('boolean', 'bool', None),
    'ASCII_Numeric_Base2': ('int', 'int', int_from_base2),
    'ASCII_Numeric_Base8': ('int', 'int', int_from_base8),
    'ASCII_Numeric_Base16': ('int', 'int', int_from_base16),
}

################################################################################
# Class Pds4TableInfo
################################################################################
class Pds4TableInfo(object):
    """The Pds4TableInfo class holds the attributes of a PDS4-labeled table."""

    def __init__(self, label_file_path, invalid={}, valid_ranges={}, table_file=None):
        """Loads a PDS4 table based on its associated label file.

        Input:
            label_file_path path to the PDS4 label file
            invalid         an optional dictionary keyed by column name. The
                            returned value must be a list or set of values that
                            are to be treated as invalid, missing or unknown.
            valid_ranges    an optional dictionary keyed by column name. The
                            returned value must be a tuple or list containing
                            the minimum and maximum numeric values in that
                            column.
            table_file      specify a table file to be read, if the provided table
                            doesn't exist in the label, an error will be raised.
        """

        # Parse PDS4 label, store the label dictionary from the pds4_tools Label object
        lbl = Label.from_file(label_file_path)
        lbl_dict = lbl.to_dict()
        self.label = lbl_dict

        # Get the file area (table file) info from the label dictionary
        file_area = None
        for prod_tag, file_area_tag in PDS4_PRODUCT_TO_FILE_AREA_TAGS_MAPPING.items():
            if prod_tag in lbl_dict.keys():
                prod_component = lbl_dict[prod_tag]
                if isinstance(file_area_tag, str):
                    file_area = prod_component[file_area_tag]
                elif isinstance(file_area_tag, tuple):
                    for tag in file_area_tag:
                        if tag in prod_component.keys():
                            file_area = prod_component[tag]
                            break
            if file_area:
                break

        self.table_file_name = None

        # The label file points to one table file
        if isinstance(file_area, dict):
            try:
                self.table_file_name = file_area['File']['file_name']
                self.table_file_li = [self.table_file_name]
            except:
                raise ValueError('Table file name was not found in PDS4 label')

            if table_file is not None and table_file not in self.table_file_li:
                raise ValueError("The provided table file name doesn't match the one" +
                                 'in the label. The label contains one table file ' +
                                 f'{self.table_file_name}')
        # The label file points to multiple table files
        elif isinstance(file_area, list):
            try:
                table_name_li = [f['File']['file_name'] for f in file_area]
            except:
                raise ValueError('Table file name was not found in PDS4 label')

            if table_file is None or table_file not in table_name_li:
                raise ValueError(f"The table file name '{table_file}' doesn't exist. " +
                                 f'The label contains {len(table_name_li)} table ' +
                                 f'files: {table_name_li}')
            else:
                self.table_file_name = table_file
                self.table_file_li = table_name_li
                # specify the table file that we want to read
                idx = table_name_li.index(table_file)
                file_area = file_area[idx]
        # The label file has no table file info
        else:
            raise ValueError(f'{label_file_path} does not contain any table file info.' )

        try:
            self.header_bytes = int(file_area['Header']['object_length'])
        except KeyError:
            # Some tables don't have header
            self.header_bytes = 0

        # Get the table/record/field info by searching the tags in the file area
        table_area = None
        record_area = None
        columns = None
        for table_tag, rec_field_tags in PDS4_TABLE_TO_RECORD_FIELD_TAGS_MAPPING.items():
            if table_tag in file_area.keys():
                record_tag, field_tag = rec_field_tags
                table_area = file_area[table_tag]
                record_area = table_area[record_tag]
                columns = record_area[field_tag]
                break

        if table_area is None or record_area is None or columns is None:
            raise ValueError(f'Missing the table/record/field info in {label_file_path}')

        self.rows = int(table_area['records'])
        self.columns = int(record_area['fields'])

        try:
            # for the table with fixed row length
            self.row_bytes = int(record_area['record_length'])
            self.fixed_length_row = True
            self.field_delimiter = None
        except:
            # for the case like .csv table, row length is not used
            self.row_bytes = int(record_area['maximum_record_length'])
            self.fixed_length_row = False
            self.field_delimiter = PDS4_FIELD_DELIMITER[table_area['field_delimiter']]

        # Save the key info about each column in a list and a dictionary
        self.column_info_list = []
        self.column_info_dict = {}

        # Construct the dtype0 dictionary
        self.dtype0 = {'crlf': ('|S2', self.row_bytes-2)}

        default_invalid = set(invalid.get('default', []))

        # Check all the column names, append the suffix _{num} to the duplicated names
        colname = defaultdict(list)
        for idx, col in enumerate(columns):
            name = col['name']
            colname[name].append(idx)

        for name, idx_li in colname.items():
            # append _{num} if there are duplicated names
            if len(idx_li) > 1:
                num = 1
                for i in idx_li:
                    columns[i]['name'] += f'_{num}'
                    num += 1

        for col in columns:
            name = col['name']
            field_num = int(col['field_number'])

            pdscol = Pds4ColumnInfo(col, field_num,
                                    invalid = invalid.get(name, default_invalid),
                                    valid_range = valid_ranges.get(name, None))

            self.column_info_list.append(pdscol)
            self.column_info_dict[pdscol.name] = pdscol
            self.dtype0[pdscol.name] = pdscol.dtype0


        self.table_file_path = os.path.join(os.path.dirname(label_file_path),
                                            self.table_file_name)


################################################################################
# class Pds4ColumnInfo
################################################################################

class Pds4ColumnInfo(object):
    """The Pds4ColumnInfo class holds the attributes of one column in a PDS4
    label."""

    def __init__(self, node_dict, column_no, invalid=set(), valid_range=None):
        """Constructor for a Pds4Column.

        Input:
            node_dict   the dictionary associated with the column info obtained
                        from pds4_tools Label object.
            column_no   the index number of this column, starting at zero.
            invalid     an optional set of discrete values that are to be
                        treated as invalid, missing or unknown.
            valid_range an optional tuple or list identifying the lower and
                        upper limits of the valid range for a numeric column.
        """

        self.name = node_dict['name']
        self.colno = column_no

        try:
            self.start_byte = int(node_dict['field_location'])
            self.bytes      = int(node_dict['field_length'])
        except:
            # For .csv table, each column length is not fixed (row is not fixed), so
            # we don't have these info.
            self.start_byte = None
            self.bytes = None

        # Handle the case where one column stores multiple items, like EXPECTED_MAXIMUM
        # in casssini iss cruise
        self.description = node_dict.get('description', '')
        items = 1
        if '-valued' in self.description:
            try:
                items = int(re.match(r'.*(\d)-valued.*', self.description.strip())[1])
            except:
                items = 1

        if self.start_byte is not None and self.bytes is not None:
            self.items = node_dict.get('ITEMS', items)
            item_bytes = int((self.bytes-items+1)/items)

            self.item_bytes = node_dict.get('ITEM_BYTES', item_bytes)
            self.item_offset = node_dict.get('ITEM_OFFSET', item_bytes+1)

            # Define dtype0 to isolate each column in a record
            self.dtype0 = ('S' + str(self.bytes), self.start_byte - 1)

            # Define dtype1 as a list of dtypes needed to isolate each item
            if self.items == 1:
                self.dtype1 = None
            else:
                self.dtype1 = {}
                byte0 = 0
                for i in range(self.items):
                    self.dtype1['item_' + str(i)] = ('S' + str(self.item_bytes),
                                                    byte0)
                    byte0 += self.item_offset
        else:
            self.dtype0 = None
            self.dtype1 = None

        # PDS4 TODO: review the data type conversion
        # Define dtype2 as the intended dtype of the values in the column
        self.data_type = node_dict['data_type']
        # Convert PDS4 data_type
        try:
            (self.data_type,
             self.dtype2,
             self.scalar_func) = PDS4_CHR_DATA_TYPE_MAPPING[self.data_type]
        except:
            raise ValueError('unsupported data type: ' + self.data_type)

        # Handle the case like "START_TIME" with ASCII_String instead of ASCII_Time as
        # the data type
        if self.name.endswith("_TIME") or self.name.endswith("_DATE"):
            self.data_type = "time"
            self.dtype2 = 'S'
            self.scalar_func = tai_from_iso

        # Identify validity criteria
        invalid_set = set()
        if valid_range is not None:
            self.valid_range = valid_range
        else:
            valid_max = None
            valid_min = None
            # Search for 'Special_Constants' tag, if it exists, get the invalid values
            # from tags in PDS4_SPECIAL_CONSTANTS_TAGS and store them in invalid_set
            if 'Special_Constants' in node_dict.keys():
                special_const_area = node_dict['Special_Constants']
                for invalid_tag in PDS4_SPECIAL_CONSTANTS_TAGS:
                    invalid_val = special_const_area.get(invalid_tag, None)
                    if invalid_val:
                        if self.scalar_func:
                            try:
                                invalid_val = self.scalar_func(invalid_val)
                            except:
                                # if the invalid value can't be converted, we will keep
                                # its original value and data type
                                invalid_val = invalid_val

                        invalid_set.add(invalid_val)

                valid_max = special_const_area.get('valid_maximum', None)
                valid_min = special_const_area.get('valid_minimum', None)

                valid_max = self.scalar_func(valid_max) if valid_max else None
                valid_min = self.scalar_func(valid_min) if valid_min else None

            self.valid_range = (valid_min, valid_max) if valid_min or valid_max else None

        if isinstance(invalid, (numbers.Real,) + STRING_TYPES):
            invalid_set |= set([invalid])
        else:
            invalid_set |= invalid

        self.invalid_values = invalid_set



################################################################################
# UNIT TESTS
################################################################################

from pdstable import *

import unittest

class Test_Pds4Table(unittest.TestCase):

  def runTest(self):

    INDEX_PATH = 'test_files/uranus_occultation_ring_fit_rfrench_20201201.xml'
    #######################################################################
    # Test csv table
    #######################################################################
    CSV_TABLE_FILE_NAME = 'uranus_occultation_ring_fit_rfrench_input_stars_20201201.csv'

    test_csv_table_basic = PdsTable(label_file=INDEX_PATH,
                                    table_file=CSV_TABLE_FILE_NAME)

    # Test strings
    test_star_names = test_csv_table_basic.column_values['Star Name']
    star_name_test_set = np.array(['Bper', 'SSgr', 'U0', 'U0201'])
    self.assertTrue(np.all(star_name_test_set == test_star_names[:4]))

    # Test ints
    test_star_num = test_csv_table_basic.column_values['Star Number']
    star_num_test_set = np.array([3, 8, 12, 16])
    self.assertTrue(np.all(star_num_test_set == test_star_num[:4]))

    # Test floats
    test_ra = test_csv_table_basic.column_values['RA(ICRS)']
    ra_test_set = np.array([47.04220716,
                            283.8163196 ,
                            219.5492129 ,
                            330.1143053])
    self.assertTrue(np.all(ra_test_set == test_ra[:4]))

    # Test dicts_by_row()
    rowdict = test_csv_table_basic.dicts_by_row()
    for i in range(4):
        self.assertEqual(rowdict[i]['RA(ICRS)'], ra_test_set[i])

    ######################################################################################
    # Test PdsTable instantiation without specifying a valid table name if multiple tables
    # are available
    ######################################################################################
    table_files = ("['uranus_occultation_ring_fit_rfrench_20201201.tab', " +
        "'uranus_occultation_ring_fit_rfrench_20201201.txt', " +
        "'uranus_occultation_ring_fit_rfrench_input_data_20201201.tab', " +
        "'uranus_occultation_ring_fit_rfrench_input_events_20201201.tab', " +
        "'uranus_occultation_ring_fit_rfrench_input_observatories_20201201.tab', " +
        "'uranus_occultation_ring_fit_rfrench_input_stars_20201201.csv']")
    try:
        test_csv_table_basic = PdsTable(label_file=INDEX_PATH)
    except ValueError as e:
        self.assertIn(table_files, str(e),
                      f'"{table_files}" NOT in error messages: "{str(e)}"')

    try:
        test_csv_table_basic = PdsTable(label_file=INDEX_PATH, table_file='xxx')
    except ValueError as e:
        self.assertIn(table_files, str(e),
                      f'"{table_files}" NOT in error messages: "{str(e)}"')

    ########################################################################
    # Row lookups
    # No File Specification or Bundle Name in .csv table, so return -1
    ########################################################################
    # File Specification
    self.assertEqual(test_csv_table_basic.filespec_column_index(), -1)
    # Bundle Name
    self.assertEqual(test_csv_table_basic.volume_column_index(), -1)

    ########################################################################
    # Row ranges
    # Can't specify row range since rows are not fixed length
    ########################################################################
    error_msg = 'cannot specify row range for the table without fixed length rows'
    try:
        partial_table = PdsTable(label_file=INDEX_PATH,
                                 row_range=(2,4),
                                 table_file=CSV_TABLE_FILE_NAME)
    except ValueError as e:
        self.assertIn(error_msg, str(e),
                      f'"{error_msg}" NOT in error messages: "{str(e)}"')

    # PDS4 TODO: Add tests for invalids & replacements

    #######################################################################
    # Test tab table
    #######################################################################
    TAB_TABLE_FILE_NAME = 'uranus_occultation_ring_fit_rfrench_20201201.tab'
    test_tab_table_basic = PdsTable(label_file=INDEX_PATH,
                                    table_file=TAB_TABLE_FILE_NAME)

    # Test strings
    test_ring_names = test_tab_table_basic.column_values['Ring name']
    ring_name_test_set = np.array(['six', 'five', 'four', 'alpha'])
    self.assertTrue(np.all(ring_name_test_set == test_ring_names[0:4]))

    # Test ints
    test_wavenum = test_tab_table_basic.column_values['Wavenumber']
    wavenum_test_set = np.array([-999, -999, -999, -999])
    self.assertTrue(np.all(wavenum_test_set == test_wavenum[:4]))

    # Test floats
    test_semimajor_axis = test_tab_table_basic.column_values['Semimajor axis']
    semimajor_axis_test_set = np.array([4.1837319048797E+04,
                                      4.2235094301041E+04,
                                      4.2571302273527E+04,
                                      4.4718670266706E+04])

    self.assertTrue(np.all(semimajor_axis_test_set == test_semimajor_axis[:4]))

    # Test dicts_by_row()
    rowdict = test_tab_table_basic.dicts_by_row()
    for i in range(4):
        self.assertEqual(rowdict[i]['Semimajor axis'], semimajor_axis_test_set[i])

    ########################################################################
    # Row lookups, no file spec or bundle name in this table, so return -1
    ########################################################################
    # File Specification
    self.assertEqual(test_tab_table_basic.filespec_column_index(), -1)
    # Bundle Name
    self.assertEqual(test_tab_table_basic.volume_column_index(), -1)

    ####################################
    # Row ranges
    ####################################

    partial_table = PdsTable(label_file=INDEX_PATH,
                             row_range=(2,4),
                             table_file=TAB_TABLE_FILE_NAME)
    self.assertEqual(partial_table.rows, 2)

    self.assertEqual(partial_table.filespec_column_index(), -1)
    self.assertEqual(partial_table.volume_column_index(), -1)

    self.assertEqual(partial_table.find_row_index(**{'Ring name': 'four'}), 0)
    self.assertEqual(partial_table.find_row_index(**{'Ring name': 'alpha'}), 1)


    ####################################
    # PdsLabel input option
    ####################################
    # For PDS4, we store the label dictionary in .lable instead of pdsparser.PdsLabel
    # instance, therefore we use "==" here instead of "is"
    test = PdsTable(label_file=INDEX_PATH,
                    label_contents=partial_table.pdslabel,
                    table_file=TAB_TABLE_FILE_NAME)
    self.assertTrue(test.pdslabel == partial_table.pdslabel)


    ####################################
    # Invalids
    ####################################
    # A dictionary stores the invalid results of each column at different rows in table:
    # uranus_occultation_ring_fit_rfrench_20201201.tab
    # key is the column name, and value is the number of invalid results in that column.
    cols_with_invalid_results = {
        'Ring name': 0,
        'Semimajor axis': 0,
        'Semimajor axis uncertainty': 0,
        'Eccentricity': 0,
        'Eccentricity uncertainty': 0,
        'Periapse longitude': 0,
        'Periapse uncertainty': 0,
        'Periapse precession rate': 0,
        'Periapse precession rate uncertainty': 1,
        'Periapse precession rate method': 0,
        'Inclination': 0,
        'Inclination uncertainty': 0,
        'Node longitude': 0,
        'Node uncertainty': 0,
        'Nodal regression rate': 0,
        'Nodal regression rate uncertainty': 6,
        'Nodal regression rate method': 0,
        'Wavenumber': 6,
        'Normal mode amplitude': 6,
        'Normal mode amplitude uncertainty': 6,
        'Normal mode phase': 6,
        'Normal mode phase uncertainty': 6,
        'Normal mode pattern speed': 6,
        'Normal mode pattern speed uncertainty': 6,
        'Number of points (Npts)': 0,
        'RMS': 0,
    }

    test_table = PdsTable(label_file=INDEX_PATH, table_file=TAB_TABLE_FILE_NAME)

    rowdict = test_table.dicts_by_row()
    for key in test_table.get_keys():
        if key.endswith('_mask'): continue
        rowmasks = test_table.get_column_mask(key)

        if cols_with_invalid_results[key] != 0:
            self.assertTrue(np.any(rowmasks))
        else:
            self.assertFalse(np.any(rowmasks))
        self.assertTrue(isinstance(rowmasks, np.ndarray))

    # rowdict = test_table.dicts_by_row()
    for key in test_table.get_keys():
        if key.endswith('_mask'): continue

        rowvals = test_table.get_column(key)
        if np.shape(rowvals[0]) != (): continue

        rowmask = test_table.get_column_mask(key)
        # print('xxxxx')
        # print(f'key: {key}')
        # print(rowvals)
        # print(rowvals.dtype.kind)

        if cols_with_invalid_results[key] != 0:
            if rowvals.dtype.kind == 'i':
                # Wavenumber
                countv = np.sum(rowvals == -999)
                self.assertEqual(countv, cols_with_invalid_results[key])
            else:
                # Rest of columns that have invalid values
                countv = np.sum(rowvals == -9.99e+99)
                self.assertEqual(countv, cols_with_invalid_results[key])
        else:
            # columns without invalid values in the table
            countv = 0

        countm = np.sum(rowmask)
        self.assertEqual(countv, countm)

    # 3.0377050292398E-04 appear 4 times in column 'Inclination'
    test_table = PdsTable(label_file=INDEX_PATH,
                          invalid={'default': [3.0377050292398E-04]},
                          table_file=TAB_TABLE_FILE_NAME)
    key = 'Inclination'
    rowmask = test_table.get_column_mask(key)
    self.assertEqual(np.sum(rowmask), 4)

    test_table = PdsTable(label_file=INDEX_PATH,
                          invalid={'default': [3.0377050292398E-04],
                                   key: 3.0377050292398E-04},
                          table_file=TAB_TABLE_FILE_NAME)

    rowmask = test_table.get_column_mask(key)
    self.assertEqual(np.sum(rowmask), 4)


    ####################################
    # Replacements
    ####################################

    # Replacement as a number
    key = 'Inclination'
    test_table = PdsTable(label_file=INDEX_PATH,
                          invalid={'default': [3.0377050292398E-04, -1.e32]},
                          replacements={key: {3.0377050292398E-04: -1.e32}},
                          table_file=TAB_TABLE_FILE_NAME)

    # replace 3.0377050292398E-04 with -1.e32
    rowvals = test_table.get_column(key)
    self.assertEqual(np.sum(rowvals == 3.0377050292398E-04), 0)
    self.assertEqual(np.sum(rowvals == -1.e32), 4)

    # still 4 invalid values because we put -1.e32 in invalid
    rowmask = test_table.get_column_mask(key)
    self.assertEqual(np.sum(rowmask), 4)



    # Replacement as a string
    test_table = PdsTable(label_file=INDEX_PATH,
                          invalid={'default': [3.0377050292398E-04, -1.e32]},
                          replacements={key: {'  3.0377050292398E-04': '  -1.e32'}},
                          table_file=TAB_TABLE_FILE_NAME)

    rowvals = test_table.get_column(key)
    self.assertEqual(np.sum(rowvals == 3.0377050292398E-04), 0)
    self.assertEqual(np.sum(rowvals == -1.e32), 4)

    rowmask = test_table.get_column_mask(key)
    self.assertEqual(np.sum(rowmask), 4)

    # Replacement via a callback
    def test_callback_as_str(arg):
        if arg.strip() == '3.0377050292398E-04': return '-1e32'
        return arg

    test_table = PdsTable(label_file=INDEX_PATH,
                          invalid={'default': [3.0377050292398E-04, -1.e32]},
                          callbacks={key: test_callback_as_str},
                          table_file=TAB_TABLE_FILE_NAME)

    rowvals = test_table.get_column(key)
    self.assertEqual(np.sum(rowvals == 3.0377050292398E-04), 0)
    self.assertEqual(np.sum(rowvals == -1.e32), 4)

    rowmask = test_table.get_column_mask(key)
    self.assertEqual(np.sum(rowmask), 4)



##########################################################################################
# pdstable/pds3table.py
# Store Pds3TableInfo and Pds3ColumnInfo
##########################################################################################
import julian
import numbers
import numpy as np
import os
from pdsparser import Pds3Label
import warnings

PDS3_VOLUME_COLNAMES = (
    'VOLUME_ID',
    'VOLUME ID',
    'VOLUME_NAME',
    'VOLUME NAME'
)

# This is an exhaustive tuple of string-like types
STRING_TYPES = (str, bytes, bytearray, np.str_, np.bytes_)

# Needed because the default value of strip is False
def tai_from_iso(string):
    return julian.tai_from_iso(string, strip=True)


################################################################################
# Class Pds3TableInfo
################################################################################
class Pds3TableInfo(object):
    """The Pds3TableInfo class holds the attributes of a PDS3-labeled table."""

    def __init__(self, label_file_path, label_list=None, invalid={},
                       valid_ranges={}, label_method='strict'):
        """Loads a PDS table based on its associated label file.

        Input:
            label_file_path path to the label file
            label_list      an option to override the parsing of the label.
                            If this is a list, it is interpreted as containing
                            all the records of the PDS label, in which case the
                            overrides the contents of the label file.
                            Alternatively, this can be a Pds3Label object that
                            was already parsed.
            invalid         an optional dictionary keyed by column name. The
                            returned value must be a list or set of values that
                            are to be treated as invalid, missing or unknown.
            valid_ranges    an optional dictionary keyed by column name. The
                            returned value must be a tuple or list containing
                            the minimum and maximum numeric values in that
                            column.
            label_method    the method to use to parse the label. Valid values
                            are 'strict' (default) or 'fast'. The 'fast' method
                            is faster but may not be as accurate.
        """

        # Parse the label
        if isinstance(label_list, (Pds3Label, dict)):
            self.label = label_list
        elif label_list:
            self.label = Pds3Label(label_list, method=label_method)
        else:
            self.label = Pds3Label(label_file_path, method=label_method)

        # Get the basic file info...
        if self.label["RECORD_TYPE"] != "FIXED_LENGTH":
            raise IOError('PDS table does not contain fixed-length records')
        else:
            # PDS3 table has fixed length rows
            self.fixed_length_row = True

        # Find the pointer to the table file
        # Confirm that the value is a PdsSimplePointer
        self.table_file_name = None
        for key, value in self.label.items():
            if key[0] == "^" and key.endswith('TABLE'):
                self.table_file_name = value
                if key + '_offset' in self.label:
                    msg = ("Table file pointer " + self.label[key + '_fmt'] +
                           " is not a Simple Pointer and isn't fully "+
                           "supported")
                    warnings.warn(msg)
                else:
                    self.table_file_name = value
                break

        if self.table_file_name is None:
            raise IOError("Pointer to a data file was not found in PDS label")

        # Locate the root of the table object
        table_dict = self.label[key[1:]]

        # Save key info about the table
        interchange_format = (table_dict.get("INTERCHANGE_FORMAT", '')
                              or table_dict["INTERCHANGE_FORMAT_1"])
        if interchange_format != "ASCII":
            raise IOError('PDS table is not in ASCII format')

        self.rows = table_dict["ROWS"]
        self.columns = table_dict["COLUMNS"]
        self.row_bytes = table_dict["ROW_BYTES"]

        # Save the key info about each column in a list and a dictionary
        self.column_info_list = []
        self.column_info_dict = {}

        # Construct the dtype0 dictionary
        self.dtype0 = {'crlf': ('|S2', self.row_bytes-2)}

        default_invalid = set(invalid.get("default", []))
        counter = 0
        for key, column_dict in table_dict.items():
            if not isinstance(column_dict, dict):
                continue
            if column_dict['OBJECT'] == "COLUMN":
                name = column_dict["NAME"]
                pdscol = Pds3ColumnInfo(column_dict, counter,
                                        invalid = invalid.get(name, default_invalid),
                                        valid_range = valid_ranges.get(name, None))
                counter += 1

                if name in self.column_info_dict:
                    raise ValueError('duplicated column name: ' + name)

                self.column_info_list.append(pdscol)
                self.column_info_dict[pdscol.name] = pdscol
                self.dtype0[pdscol.name] = pdscol.dtype0

        # Fill in the complete table file name
        self.table_file_path = os.path.join(os.path.dirname(label_file_path),
                                            self.table_file_name)

################################################################################
# class Pds3ColumnInfo
################################################################################

class Pds3ColumnInfo(object):
    """The PdsColumnInfo class holds the attributes of one column in a PDS
    label."""

    def __init__(self, node_dict, column_no, invalid=set(), valid_range=None):
        """Constructor for a PdsColumn.

        Input:
            node_dict   the dictionary associated with the pdsparser.PdsNode
                        object defining the column.
            column_no   the index number of this column, starting at zero.
            invalid     an optional set of discrete values that are to be
                        treated as invalid, missing or unknown.
            valid_range an optional tuple or list identifying the lower and
                        upper limits of the valid range for a numeric column.
        """

        self.name = node_dict["NAME"]
        self.colno = column_no

        self.start_byte = node_dict["START_BYTE"]
        self.bytes      = node_dict["BYTES"]

        self.items = node_dict.get("ITEMS", 1)
        self.item_bytes = node_dict.get("ITEM_BYTES", self.bytes)
        self.item_offset = node_dict.get("ITEM_OFFSET", self.bytes)

        # Define dtype0 to isolate each column in a record
        self.dtype0 = ("S" + str(self.bytes), self.start_byte - 1)

        # Define dtype1 as a list of dtypes needed to isolate each item
        if self.items == 1:
            self.dtype1 = None
        else:
            self.dtype1 = {}
            byte0 = 0
            for i in range(self.items):
                self.dtype1["item_" + str(i)] = ("S" + str(self.item_bytes),
                                                 byte0)
                byte0 += self.item_offset

        # Define dtype2 as the intended dtype of the values in the column
        self.data_type = node_dict["DATA_TYPE"]
        if "INTEGER" in self.data_type:
            self.data_type = "int"
            self.dtype2 = "int"
            self.scalar_func = int
        elif "REAL" in self.data_type:
            self.data_type = "float"
            self.dtype2 = "float"
            self.scalar_func = float
        elif ("TIME" in self.data_type or "DATE" in self.data_type or
              self.name.endswith("_TIME") or self.name.endswith("_DATE")):
            self.data_type = "time"
            self.dtype2 = 'S'
            self.scalar_func = tai_from_iso
        elif "CHAR" in self.data_type:
            self.data_type = "string"
            self.dtype2 = 'U'
            self.scalar_func = None
        else:
            raise IOError("unsupported data type: " + self.data_type)

        # Identify validity criteria
        self.valid_range = valid_range or node_dict.get("VALID_RANGE", None)

        if isinstance(invalid, (numbers.Real,) + STRING_TYPES):
            invalid = set([invalid])

        self.invalid_values = set(invalid)

        self.invalid_values.add(node_dict.get("INVALID_CONSTANT"       , None))
        self.invalid_values.add(node_dict.get("MISSING_CONSTANT"       , None))
        self.invalid_values.add(node_dict.get("UNKNOWN_CONSTANT"       , None))
        self.invalid_values.add(node_dict.get("NOT_APPLICABLE_CONSTANT", None))
        self.invalid_values.add(node_dict.get("NULL_CONSTANT"          , None))
        self.invalid_values.add(node_dict.get("INVALID"                , None))
        self.invalid_values.add(node_dict.get("MISSING"                , None))
        self.invalid_values.add(node_dict.get("UNKNOWN"                , None))
        self.invalid_values.add(node_dict.get("NOT_APPLICABLE"         , None))
        self.invalid_values.add(node_dict.get("NULL"                   , None))
        self.invalid_values -= {None}



################################################################################
# UNIT TESTS
################################################################################

from pdstable import *

import unittest

class Test_Pds4Table(unittest.TestCase):

  def runTest(self):

    INDEX_PATH = 'test_files/cassini_iss_cruise_index_edited.xml'

    test_table_basic = PdsTable(INDEX_PATH)

    # Test strings
    test_file_names = test_table_basic.column_values['File Name']
    file_name_test_set = np.array(['1294561143w.img',
                                   '1294561202w.img',
                                   '1294561261w.img',
                                   '1294561333w.img'])

    self.assertTrue(np.all(file_name_test_set == test_file_names[:4]))

    # Test ints
    test_cmd_seq_num_idx = test_table_basic.column_values['COMMAND_SEQUENCE_NUMBER']
    cmd_seq_num_idx_test_set = np.array([65534, 65534, 65534, 65534])
    self.assertTrue(np.all(cmd_seq_num_idx_test_set == test_cmd_seq_num_idx[:4]))

    # Test floats
    test_dark_strip_mean = test_table_basic.column_values['DARK_STRIP_MEAN']
    dark_strip_mean_test_set = np.array([68.44, 68.47, 68.39, 68.47])
    self.assertTrue(np.all(dark_strip_mean_test_set == test_dark_strip_mean[:4]))

    # Test times as strings
    test_start_time_strs = test_table_basic.column_values['START_TIME']
    start_time_str_test_set = ['1999-009T08:14:21.687',
                               '1999-009T08:15:20.687',
                               '1999-009T08:16:19.687',
                               '1999-009T08:17:31.686']

    self.assertEqual(start_time_str_test_set[0], test_start_time_strs[0])
    self.assertEqual(start_time_str_test_set[1], test_start_time_strs[1])
    self.assertEqual(start_time_str_test_set[2], test_start_time_strs[2])
    self.assertEqual(start_time_str_test_set[3], test_start_time_strs[3])

    self.assertTrue(isinstance(test_start_time_strs, np.ndarray))
    self.assertTrue(isinstance(test_start_time_strs[0], np.str_))

    # Test dicts_by_row()
    rowdict = test_table_basic.dicts_by_row()
    for i in range(4):
        self.assertEqual(rowdict[i]['START_TIME'], test_start_time_strs[i])

    rowvals = test_table_basic.get_column('START_TIME')
    rowmasks = test_table_basic.get_column_mask('START_TIME')
    for i in range(10):
        self.assertEqual(rowdict[i]['START_TIME'], rowvals[i])
        self.assertFalse(rowmasks[i])

    ####################################
    # Test times as seconds (floats)
    ####################################

    test_table_secs = PdsTable(INDEX_PATH, times=['START_TIME'])

    test_start_times = test_table_secs.column_values['START_TIME']
    start_time_test_set = np.array([-30858306.313,
                                    -30858247.313,
                                    -30858188.313,
                                    -30858116.314])

    self.assertTrue(np.all(start_time_test_set == test_start_times[:4]))
    self.assertTrue(isinstance(start_time_test_set, np.ndarray))

    # Test dicts_by_row()
    rowdict = test_table_secs.dicts_by_row()
    for i in range(4):
        self.assertEqual(rowdict[i]['START_TIME'], start_time_test_set[i])

    rowvals = test_table_secs.get_column('START_TIME')
    rowmask = test_table_secs.get_column_mask('START_TIME')
    for i in range(10):
        self.assertEqual(rowdict[i]['START_TIME'], rowvals[i])
        self.assertFalse(rowmask[i])

    ####################################
    # Row lookups
    ####################################
    # FILE_SPECIFICATION_NAME
    self.assertEqual(test_table_basic.filespec_column_index(), 3)
    # VOLUME_ID
    self.assertEqual(test_table_basic.volume_column_index(), 4)
    # FILE_SPECIFICATION_NAME
    self.assertEqual(test_table_basic.find_row_index_by_volume_filespec(
            '', 'data/1294561143_1295221348/W1294561261_1.IMG'), 2)
    self.assertEqual(test_table_basic.find_row_indices_by_volume_filespec(
            '', 'data/1294561143_1295221348/W1294561261_1.IMG'), [2])
    self.assertEqual(test_table_basic.find_row_index_by_volume_filespec(
            '', 'data/1294561143_1295221348/W1294561333_1.IMG'), 3)
    self.assertEqual(test_table_basic.find_row_indices_by_volume_filespec(
            '', 'data/1294561143_1295221348/W1294561333_1.IMG'), [3])
    # VOLUME_ID & FILE_SPECIFICATION_NAME
    self.assertEqual(test_table_basic.find_row_index_by_volume_filespec(
            'COISS_1001',
            'data/1294561143_1295221348/N1294562836_1.IMG'), 15)
    self.assertEqual(test_table_basic.find_row_indices_by_volume_filespec(
            'COISS_1001',
            'data/1294561143_1295221348/N1294562836_1.IMG'), [15])


    ####################################
    # Row ranges
    ####################################

    partial_table = PdsTable(INDEX_PATH, row_range=(2,4))
    self.assertEqual(partial_table.rows, 2)

    self.assertEqual(partial_table.filespec_column_index(), 3)
    self.assertEqual(partial_table.volume_column_index(), 4)

    self.assertEqual(partial_table.find_row_index_by_volume_filespec(
            '', 'data/1294561143_1295221348/W1294561261_1.IMG'), 0)
    self.assertEqual(partial_table.find_row_indices_by_volume_filespec(
            '', 'data/1294561143_1295221348/W1294561261_1.IMG'), [0])

    self.assertEqual(partial_table.find_row_index_by_volume_filespec(
            '', 'data/1294561143_1295221348/W1294561333_1.IMG'), 1)
    self.assertEqual(partial_table.find_row_indices_by_volume_filespec(
            '', 'data/1294561143_1295221348/W1294561333_1.IMG'), [1])
    self.assertEqual(partial_table.find_row_index_by_volume_filespec(
            'COISS_1001',
            'data/1294561143_1295221348/W1294561261_1.IMG'), 0)
    self.assertEqual(partial_table.find_row_indices_by_volume_filespec(
            'COISS_1001',
            'data/1294561143_1295221348/W1294561261_1.IMG'), [0])

    ####################################
    # PdsLabel input option
    ####################################
    # For PDS4, we store the label dictionary in .lable instead of pdsparser.PdsLabel
    # instance, therefore we use "==" here instead of "is"
    test = PdsTable(INDEX_PATH, label_contents=partial_table.pdslabel)
    self.assertTrue(test.pdslabel == partial_table.pdslabel)

    # PDS4 TODO: Add tests for invalids & replacements
